# distutils: language = c
# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
import numpy as np
cimport numpy as np
cimport cython
from libc.stdio cimport printf

# Import the header files
from cyvlfeat._vl.hog cimport *
from cyvlfeat._vl.host cimport VL_FALSE


@cython.boundscheck(False)
cpdef cy_hog(float[:, :, ::1] data, int cell_size, int variant,
             int n_orientations, bint directed_polar_field,
             bint undirected_polar_field, bint bilinear_interpolation,
             bint return_channels_last_axis, bint verbose):

    cdef:
        # Python images are not transposed
        VlHog* hog = vl_hog_new(<VlHogVariant>variant,
                                n_orientations, VL_FALSE)
        int height = data.shape[0]
        int width = data.shape[1]
        int n_channels = data.shape[2]
        int out_height = 0, out_width = 0, out_n_channels = 0
        np.ndarray[float, ndim=3, mode='c'] out_array

    vl_hog_set_use_bilinear_orientation_assignments(hog,
                                                    bilinear_interpolation)

    if directed_polar_field or undirected_polar_field:
        vl_hog_put_polar_field(hog,
                               &data[0, 0, 0],  # Magnitude
                               &data[0, 0, 1],  # Angle
                               directed_polar_field,
                               width, height, cell_size)
    else:  # Assume we have an image
        vl_hog_put_image(hog, &data[0, 0, 0], width, height, n_channels,
                         cell_size)

    out_height = vl_hog_get_height(hog)
    out_width = vl_hog_get_width(hog)
    out_n_channels = vl_hog_get_dimension(hog)

    if verbose:
        printf('vl_hog: image: [%d x %d x %d]\n', height, width, n_channels)
        printf('vl_hog: descriptor: [%d x %d x %d]\n', out_height, out_width,
                                                       out_n_channels)
        printf('vl_hog: number of orientations: %d\n', n_orientations)
        printf('vl_hog: bilinear orientation assignments: %s\n',
               'yes' if bilinear_interpolation else 'no')
        printf('vl_hog: variant: %s\n',
               'DalalTriggs' if variant == VlHogVariantDalalTriggs
               else 'UOCTTI')
        printf('vl_hog: input type: %s\n',
               'DirectedPolarField' if directed_polar_field
               else ('UndirectedPolarField' if undirected_polar_field else
                     'Image'))

    # Unfortunately, writing in C-contiguous ordering implies the channels
    # should be at the front.
    out_array = np.empty((out_n_channels, out_height, out_width),
                         dtype=np.float32, order='C')

    vl_hog_extract(hog, &out_array[0, 0, 0])
    vl_hog_delete(hog)

    # Therefore, a copy is required if the channels should be returned as the
    # last axis.
    if return_channels_last_axis:
        out_array = np.transpose(out_array, [1, 2, 0]).copy()

    return out_array



from __future__ import division
from cyvlfeat.hog import hog
import numpy as np
from numpy.testing import assert_allclose
from cyvlfeat.test_util import lena

        
img = lena().astype(np.float32) / 255.
half_img = img[:, :256]


def test_hog_cell_size_32_uoctti():
    i = img.copy()
    output = hog(i, 32)
    assert output.dtype == np.float32
    assert output.shape == (16, 16, 31)
    assert_allclose(output[:2, :2, 0],
                    np.array([[0.16276041, 0.13416694],
                              [0.22367628, 0.17556792]]), rtol=1e-5)


def test_hog_cell_size_32_uoctti_4_orientations():
    i = img.copy()
    output = hog(i, 32, n_orientations=4)
    assert output.dtype == np.float32
    assert output.shape == (16, 16, 16)
    assert_allclose(output[:2, :2, 0],
                    np.array([[0.23047766, 0.17703892],
                              [0.30189356, 0.25290328]]), rtol=1e-5)


def test_hog_cell_size_32_uoctti_non_square():
    i = half_img.copy()
    output = hog(i, 32)
    assert output.dtype == np.float32
    assert output.shape == (16, 8, 31)
    assert_allclose(output[:2, :2, 0],
                    np.array([[0.16276041, 0.13416694],
                              [0.22367628, 0.17556792]]), rtol=1e-5)


def test_hog_cell_size_32_dalaltriggs():
    i = img.copy()
    output = hog(i, 32, variant='DalalTriggs')
    assert output.dtype == np.float32
    assert output.shape == (16, 16, 36)
    assert_allclose(output[:2, :2, 0],
                    np.array([[0.2, 0.2],
                              [0.2, 0.2]]))


def test_hog_cell_size_32_dalaltriggs_4_orientations():
    i = img.copy()
    output = hog(i, 32, n_orientations=4, variant='DalalTriggs')
    assert output.dtype == np.float32
    assert output.shape == (16, 16, 16)
    assert_allclose(output[:2, :2, 0],
                    np.array([[0.2, 0.2],
                              [0.2, 0.2]]))


def test_hog_cell_size_32_dalaltriggs_non_square():
    i = half_img.copy()
    output = hog(i, 32, variant='DalalTriggs')
    assert output.dtype == np.float32
    assert output.shape == (16, 8, 36)
    assert_allclose(output[:2, :2, 0],
                    np.array([[0.2, 0.2],
                              [0.2, 0.2]]), rtol=1e-5)


def test_hog_cell_size_32_dalaltriggs_bilinear_interpolation():
    i = img.copy()
    output = hog(i, 32, variant='DalalTriggs', bilinear_interpolation=True)
    assert output.dtype == np.float32
    assert output.shape == (16, 16, 36)
    assert_allclose(output[:2, -2:, 0],
                    np.array([[0.082442075, 0.13325043],
                              [0.094600961, 0.090005033]]), rtol=1e-5)



# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
from .host cimport vl_size, vl_index, vl_bool


cdef extern from "vl/hog.h":
    cdef enum VlHogVariant_:
        VlHogVariantDalalTriggs = 0,
        VlHogVariantUoctti = 1
    ctypedef VlHogVariant_ VlHogVariant

    cdef struct VlHog_:
        VlHogVariant variant
        vl_size dimension
        vl_size numOrientations
        vl_bool transposed
        vl_bool useBilinearOrientationAssigment

        # left-right flip permutation
        vl_index* permutation

        # glyphs
        float* glyphs
        vl_size glyphSize

        # helper vectors
        float* orientationX
        float* orientationY

        # buffers
        float* hog
        float* hogNorm
        vl_size hogWidth
        vl_size hogHeight
    ctypedef VlHog_ VlHog

    VlHog* vl_hog_new(VlHogVariant variant, vl_size numOrientations,
                      vl_bool transposed)
    void vl_hog_delete(VlHog* self)
    void vl_hog_process(VlHog* self, float* features, float* image,
                        vl_size width, vl_size height, vl_size numChannels,
                        vl_size cellSize)

    void vl_hog_put_image(VlHog * self, float* image, vl_size width,
                          vl_size height, vl_size numChannels, vl_size cellSize)

    void vl_hog_put_polar_field(VlHog* self, float* modulus,
                                float* angle, vl_bool directed,
                                vl_size width, vl_size height, vl_size cellSize)

    void vl_hog_extract(VlHog * self, float* features)
    vl_size vl_hog_get_height(VlHog * self)
    vl_size vl_hog_get_width(VlHog * self)


    void vl_hog_render(VlHog * self, float* image, float* features,
                       vl_size width, vl_size height)

    vl_size vl_hog_get_dimension(VlHog* self)
    vl_index* vl_hog_get_permutation(VlHog* self)
    vl_size vl_hog_get_glyph_size(VlHog* self)

    vl_bool vl_hog_get_use_bilinear_orientation_assignments(VlHog* self)
    void vl_hog_set_use_bilinear_orientation_assignments(VlHog* self,
                                                         vl_bool x)



# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
from cyvlfeat._vl.host cimport vl_bool

cdef extern from "vl/dsift.h":
    ctypedef struct VlDsiftKeypoint:
        double x  #< x coordinate */
        double y  #< y coordinate */
        double s  #< scale */
        double norm  #< SIFT descriptor norm */

# @brief Dense SIFT descriptor geometry */
    ctypedef struct VlDsiftDescriptorGeometry:
        int numBinT   #< number of orientation bins */
        int numBinX   #< number of bins along X */
        int numBinY   #< number of bins along Y */
        int binSizeX  #< size of bins along X */
        int binSizeY  #< size of bins along Y */


    ctypedef struct VlDsiftFilter:
        int imWidth             #< @internal @brief image width */
        int imHeight            #< @internal @brief image height */

        int stepX               #< frame sampling step X */
        int stepY               #< frame sampling step Y */

        int boundMinX           #< frame bounding box min X */
        int boundMinY           #< frame bounding box min Y */
        int boundMaxX           #< frame bounding box max X */
        int boundMaxY           #< frame bounding box max Y */

        # descriptor parameters */
        VlDsiftDescriptorGeometry geom

        int useFlatWindow       #< flag: whether to approximate the Gaussian window with a flat one */
        double windowSize       #< size of the Gaussian window */

        int numFrames           #< number of sampled frames */
        int descrSize           #< size of a descriptor */
        VlDsiftKeypoint *frames  #< frame buffer */
        float *descrs           #< descriptor buffer */

        int numBinAlloc         #< buffer allocated: descriptor size */
        int numFrameAlloc       #< buffer allocated: number of frames  */
        int numGradAlloc        #< buffer allocated: number of orientations */

        float **grads           #< gradient buffer */
        float *convTmp1         #< temporary buffer */
        float *convTmp2         #< temporary buffer */


    VlDsiftFilter *vl_dsift_new(int width, int height)
    VlDsiftFilter *vl_dsift_new_basic(int width, int height, int step,
                                      int binSize)
    void vl_dsift_delete(VlDsiftFilter *self)
    void vl_dsift_process(VlDsiftFilter *self, float*im)
    inline void vl_dsift_transpose_descriptor(float*dst,
                                              float*src,
                                              int numBinT,
                                              int numBinX,
                                              int numBinY)

    inline void vl_dsift_set_steps(VlDsiftFilter *self,
                                   int stepX,
                                   int stepY)
    inline void vl_dsift_set_bounds(VlDsiftFilter *self,
                                    int minX,
                                    int minY,
                                    int maxX,
                                    int maxY)
    inline void vl_dsift_set_geometry(VlDsiftFilter *self,
                                      VlDsiftDescriptorGeometry*geom)
    inline void vl_dsift_set_flat_window(VlDsiftFilter *self,
                                         vl_bool useFlatWindow)
    inline void vl_dsift_set_window_size(VlDsiftFilter *self, double windowSize)

    inline float *vl_dsift_get_descriptors(VlDsiftFilter *self)
    inline int             vl_dsift_get_descriptor_size(VlDsiftFilter *self)
    inline int             vl_dsift_get_keypoint_num(VlDsiftFilter *self)
    inline VlDsiftKeypoint *vl_dsift_get_keypoints(VlDsiftFilter *self)
    inline void            vl_dsift_get_bounds(VlDsiftFilter *self,
                                               int*minX,
                                               int*minY,
                                               int*maxX,
                                               int*maxY)
    inline void            vl_dsift_get_steps(VlDsiftFilter*self,
                                              int*stepX,
                                              int*stepY)
    inline VlDsiftDescriptorGeometry* vl_dsift_get_geometry	(	VlDsiftFilter* 	self	)
    inline vl_bool         vl_dsift_get_flat_window(VlDsiftFilter *self)
    inline double          vl_dsift_get_window_size(VlDsiftFilter *self)

    void _vl_dsift_update_buffers(VlDsiftFilter *self)



# distutils: language = c
# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
import numpy as np
import ctypes
cimport numpy as np
cimport cython

# Import the header files
from cyvlfeat._vl.host cimport VL_TYPE_FLOAT
from cyvlfeat._vl.host cimport vl_size
from cyvlfeat._vl.fisher cimport vl_fisher_encode
from cyvlfeat._vl.fisher cimport VL_FISHER_FLAG_NORMALIZED
from cyvlfeat._vl.fisher cimport VL_FISHER_FLAG_SQUARE_ROOT
from cyvlfeat._vl.fisher cimport VL_FISHER_FLAG_IMPROVED
from cyvlfeat._vl.fisher cimport VL_FISHER_FLAG_FAST


@cython.boundscheck(False)
cpdef cy_fisher(float[:, :] X,
                float[:, :] MEANS,
                float[:, :] COVARIANCES,
                float[:] PRIORS,
                bint normalized,
                bint square_root,
                bint improved,
                bint fast,
                bint verbose):

    cdef:
        vl_size n_clusters = MEANS.shape[1]
        vl_size n_dimensions = MEANS.shape[0]
        vl_size n_data = X.shape[1]
        int flags = 0

    if normalized:
        flags |= VL_FISHER_FLAG_NORMALIZED
    
    if square_root:
        flags |= VL_FISHER_FLAG_SQUARE_ROOT
        
    if improved:
        flags |= VL_FISHER_FLAG_IMPROVED
        
    if fast:
        flags |= VL_FISHER_FLAG_FAST
    
    if verbose:
        print('vl_fisher: num data:       {}\n'
              'vl_fisher: num clusters:   {}\n'
              'vl_fisher: data dimension: {}\n'
              'vl_fisher: code dimension: {}\n'
              'vl_fisher: square root:    {}\n'
              'vl_fisher: normalized:     {}\n'
              'vl_fisher: fast:           {}'.format(
            n_data, n_clusters, n_dimensions, 2 * n_clusters * n_dimensions,
            square_root, normalized, fast))

    cdef float[:] enc = np.zeros(2 * n_clusters * n_dimensions,
                                 dtype=np.float32)

    cdef int num_terms = vl_fisher_encode(&enc[0],
                                          VL_TYPE_FLOAT,
                                          &MEANS[0, 0],
                                          n_dimensions,
                                          n_clusters,
                                          &COVARIANCES[0, 0],
                                          &PRIORS[0],
                                          &X[0, 0],
                                          n_data,
                                          flags)

    if verbose:
        print('vl_fisher: sparsity of assignments: {:.2f}% '
              '({} non-negligible assignments)'.format(
            100.0 * (1.0 - <float>num_terms / (n_data * n_clusters + 1e-12)),
            num_terms))

    return np.asarray(enc)



# distutils: language = c
# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
import numpy as np
cimport numpy as np
cimport cython
from cython.operator cimport dereference as deref
from libc.stdio cimport printf
from libc.stdlib cimport qsort

# Import the header files
from cyvlfeat._vl.dsift cimport *
from cyvlfeat._vl.host cimport *
from cyvlfeat._vl.sift cimport *
from cyvlfeat._vl.mathop cimport VL_PI


@cython.boundscheck(False)
cpdef cy_dsift(np.ndarray[float, ndim=2, mode='c'] data, int[:] step,
               int[:] size, int[:] bounds, int window_size, bint norm,
               bint fast, bint float_descriptors, int[:] geometry,
               bint verbose):

    cdef:
        int num_frames = 0
        VlDsiftKeypoint *frames_array
        float *descriptors_array
        int k = 0, i = 0
        int step_x = 0, step_y = 0, min_x = 0, min_y = 0, max_x = 0, max_y = 0

        np.ndarray[float, ndim=2, mode='c'] out_descriptors
        np.ndarray[float, ndim=2, mode='c'] out_frames

        int height = data.shape[0]
        int width = data.shape[1]

        VlDsiftDescriptorGeometry geom
        VlDsiftFilter *dsift = vl_dsift_new(width, height)

        int ndims = 0
        int descriptor_index = 0
        float* linear_descriptor

    # Setup the geometry (number of bins and sizes)
    # Note the y-axis is taken as the first (zeroth) axis
    geom.numBinX = geometry[1]
    geom.numBinY = geometry[0]
    geom.numBinT = geometry[2]
    geom.binSizeX = size[1]
    geom.binSizeY = size[0]
    vl_dsift_set_geometry(dsift, &geom)

    # Set other options
    vl_dsift_set_steps(dsift, step[1], step[0])
    vl_dsift_set_bounds(dsift, bounds[1], bounds[0], bounds[3], bounds[2])
    vl_dsift_set_flat_window(dsift, fast)

    if window_size >= 0:
      vl_dsift_set_window_size(dsift, window_size)

    # Get calculated values from the dsift object
    num_frames = vl_dsift_get_keypoint_num(dsift)
    descriptor_length = vl_dsift_get_descriptor_size(dsift)
    geom = deref(vl_dsift_get_geometry(dsift))

    if verbose:
      vl_dsift_get_steps(dsift, &step_x, &step_y)
      vl_dsift_get_bounds(dsift, &min_x, &min_y, &max_x, &max_y)

      printf("vl_dsift: image size         [W, H] = [%d, %d]\n", width, height)
      printf("vl_dsift: bounds:            "
             "[minX, minY, maxX, maxY] = [%d, %d, %d, %d]\n",
             min_x, min_y, max_x, max_y)
      printf("vl_dsift: subsampling steps: stepX=%d, stepY=%d\n",
             step_x, step_y)
      printf("vl_dsift: num bins:          "
             "[numBinT, numBinX, numBinY] = [%d, %d, %d]\n",
             geom.numBinT,
             geom.numBinX,
             geom.numBinY)
      printf("vl_dsift: descriptor size:   %d\n", descriptor_length)
      printf("vl_dsift: bin sizes:         [binSizeX, binSizeY] = [%d, %d]\n",
                geom.binSizeX,
                geom.binSizeY)
      printf("vl_dsift: flat window:       %d\n", fast)
      printf("vl_dsift: window size:       "
             "%g\n", vl_dsift_get_window_size(dsift))
      printf("vl_dsift: num of features:   %d\n", num_frames)

    # Actually compute the SIFT features
    vl_dsift_process(dsift, &data[0, 0])

    # Grab the results
    frames_array = vl_dsift_get_keypoints(dsift)
    descriptors_array = vl_dsift_get_descriptors(dsift)

    # Create output arrays
    out_descriptors = np.empty((num_frames, descriptor_length),
                               dtype=np.float32, order='C')
    # Grab the pointer to the data so we can walk it linearly
    linear_descriptor = &out_descriptors[0, 0]

    # The norm is added as the third component if set
    if norm:
        ndims = 3
        out_frames = np.empty((num_frames, ndims), dtype=np.float32)
    else:
        ndims = 2
        out_frames = np.empty((num_frames, ndims), dtype=np.float32)

    # Copy results out
    for k in range(num_frames):
        out_frames[k, 0] = frames_array[k].y
        out_frames[k, 1] = frames_array[k].x

        # We have an implied / 2 in the norm, because of the clipping below
        if norm:
            out_frames[k, 2] = frames_array[k].norm

        # We don't need to transpose because our memory is in the correct
        # order already!
        for i in range(descriptor_length):
            descriptor_index = num_frames * i + k
            linear_descriptor[descriptor_index] = \
                min(512.0 * descriptors_array[descriptor_index], 255.0)

    # Clean up the allocated memory
    vl_dsift_delete(dsift)

    if float_descriptors:
        return out_frames, out_descriptors
    else:
        return out_frames, out_descriptors.astype(np.uint8)


cdef int korder(const void *a, const void *b) nogil:
    cdef float x = (<float*> a)[2] - (<float*> b)[2]
    if x < 0: return -1
    if x > 0: return +1
    return 0

cdef inline void transpose_descriptor(vl_sift_pix* dst, vl_sift_pix* src) nogil:
    cdef:
        int BO = 8  # number of orientation bins
        int BP = 4  # number of spatial bins
        int i = 0, j = 0, t = 0, jp = 0, o = 0, op = 0

    for j in range(BP):
        jp = BP - 1 - j
        for i in range(BP):
            o  = BO * i + BP * BO * j
            op = BO * i + BP * BO * jp
            dst[op] = src[o]
            for t in range(1, BO):
                dst [BO - t + op] = src [t + o]


@cython.boundscheck(False)
cpdef cy_sift(np.ndarray[float, ndim=2, mode='c'] data, int n_octaves,
              int n_levels, int first_octave, int peak_threshold,
              int edge_threshold, float norm_threshold, int magnification,
              int window_size, float[:, :] frames, bint force_orientations,
              bint float_descriptors, bint compute_descriptor, bint verbose):

    cdef:
        bint is_first_octave = True
        int n_user_keypoints = 0, total_keypoints = 0, reserved = 0, i = 0, j = 0, q = 0
        int height = data.shape[0]
        int width = data.shape[1]
        float *user_keypoints_arr
        VlSiftFilt *filt = vl_sift_new(width, height, n_octaves, n_levels,
                                       first_octave)

        # Create empty 2D output arrays
        np.ndarray[float, ndim=2, mode='c'] out_descriptors = np.empty(
            (0, 128), dtype=np.float32, order='C')
        np.ndarray[float, ndim=2, mode='c'] out_frames = np.empty(
            (0, 4), dtype=np.float32, order='C')

        float *flat_descriptors = &out_descriptors[0, 0]
        float *flat_out_frames = &out_frames[0, 0]

        int is_octaves_complete = 0
        VlSiftKeypoint *keypoints
        int n_keypoints = 0

        double[4] angles
        int n_angles = 0
        VlSiftKeypoint *curr_keypoint
        VlSiftKeypoint ik

        vl_sift_pix[128] single_descriptor_arr
        vl_sift_pix[128] single_descriptor_arr_t

        bint user_specified_frames = False

    user_specified_frames = frames is not None
    if user_specified_frames:
        n_user_keypoints = frames.shape[0]
        user_keypoints_arr = &frames[0, 0]
        # Ensure frames array is sorted by increasing scale
        qsort(user_keypoints_arr, n_user_keypoints,
              4 * sizeof(float), korder)

    if peak_threshold  >= 0: vl_sift_set_peak_thresh(filt, peak_threshold)
    if edge_threshold  >= 0: vl_sift_set_edge_thresh(filt, edge_threshold)
    if norm_threshold  >= 0: vl_sift_set_norm_thresh(filt, norm_threshold)
    if magnification   >= 0: vl_sift_set_magnif(filt, magnification)
    if window_size     >= 0: vl_sift_set_window_size(filt, window_size)

    if verbose:
        printf("vl_sift: filter settings:\n")
        printf("vl_sift:   octaves      (O)      = %d\n", vl_sift_get_noctaves(filt))
        printf("vl_sift:   levels       (S)      = %d\n", vl_sift_get_nlevels(filt))
        printf("vl_sift:   first octave (o_min)  = %d\n", vl_sift_get_octave_first(filt))
        printf("vl_sift:   edge thresh           = %g\n", vl_sift_get_edge_thresh(filt))
        printf("vl_sift:   peak thresh           = %g\n", vl_sift_get_peak_thresh(filt))
        printf("vl_sift:   norm thresh           = %g\n", vl_sift_get_norm_thresh(filt))
        printf("vl_sift:   window size           = %g\n", vl_sift_get_window_size(filt))
        printf("vl_sift:   float descriptor      = %d\n", float_descriptors)

        printf("vl_sift: will source frames? yes (%d read)\n"
               if user_specified_frames
               else "vl_sift: will source frames? no\n", n_user_keypoints)
        printf("vl_sift: will force orientations? %d\n", force_orientations)


    if user_specified_frames:
        # If we have specified the frames, we know how many frames
        # will be calculated, so we can skip the dynamic reallocation
        # which is normally done inside the keypoints loop
        out_frames = np.resize(out_frames, (n_user_keypoints, 4))
        flat_out_frames = &out_frames[0, 0]
        # Similar for the descriptors, if necessary
        if compute_descriptor:
            out_descriptors = np.resize(out_descriptors,
                                        (n_user_keypoints, 128))
            flat_descriptors = &out_descriptors[0, 0]

    # Process each octave
    while True:
        if verbose:
            printf("vl_sift: processing octave %d\n",
                   vl_sift_get_octave_index(filt))

        # Calculate the GSS for the next octave ....................
        if is_first_octave:
            is_octaves_complete = vl_sift_process_first_octave(filt,
                                                               &data[0, 0])
            is_first_octave = False
        else:
            is_octaves_complete = vl_sift_process_next_octave(filt)

        if is_octaves_complete:
            break

        if verbose:
            printf("vl_sift: GSS octave %d computed\n",
                   vl_sift_get_octave_index(filt))

        # Run detector .............................................
        if not user_specified_frames:
            i = 0
            vl_sift_detect(filt)

            keypoints = vl_sift_get_keypoints(filt)
            n_keypoints = vl_sift_get_nkeypoints(filt)

            if verbose:
                printf("vl_sift: detected %d (unoriented) keypoints\n",
                       n_keypoints)
        else:
            n_keypoints = n_user_keypoints

        # For each keypoint
        while i < n_keypoints:
            # Obtain keypoint orientations
            if user_specified_frames:
                vl_sift_keypoint_init(filt, &ik,
                                      user_keypoints_arr[4 * i + 1],
                                      user_keypoints_arr[4 * i + 0],
                                      user_keypoints_arr[4 * i + 2])

                if ik.o != vl_sift_get_octave_index(filt):
                    break

                curr_keypoint = &ik

                # Optionally force computation of orientations
                if force_orientations:
                    n_angles = vl_sift_calc_keypoint_orientations(filt, angles,
                                                                  curr_keypoint)
                else:
                    angles[0] = (VL_PI / 2) - user_keypoints_arr[4 * i + 3]
                    n_angles  = 1
            else:
                # This is equivalent to &keypoints[i] - just get the pointer
                # to the i'th element.
                curr_keypoint = keypoints + i
                n_angles = vl_sift_calc_keypoint_orientations(filt, angles,
                                                              curr_keypoint)

            # For each orientation
            for q in range(n_angles):
                if compute_descriptor:
                    vl_sift_calc_keypoint_descriptor(filt,
                                                     single_descriptor_arr,
                                                     curr_keypoint, angles[q])
                    transpose_descriptor(single_descriptor_arr_t,
                                         single_descriptor_arr)

                # Dynamically reallocate the output arrays so that they can
                # fit all the keypoints being requested.
                if reserved < total_keypoints + 1:
                    reserved += 2 * n_keypoints

                    out_frames = np.resize(out_frames, (reserved, 4))
                    flat_out_frames = &out_frames[0, 0]

                    if compute_descriptor:
                        out_descriptors = np.resize(out_descriptors,
                                                    (reserved, 128))
                        flat_descriptors = &out_descriptors[0, 0]

                # Notice that this method will give different results
                # from MATLAB because MATLAB actually runs on the
                # transpose of the image due to it's fortran ordering!
                flat_out_frames[total_keypoints * 4 + 0] = curr_keypoint.y
                flat_out_frames[total_keypoints * 4 + 1] = curr_keypoint.x
                flat_out_frames[total_keypoints * 4 + 2] = curr_keypoint.sigma
                flat_out_frames[total_keypoints * 4 + 3] = (VL_PI / 2) - angles[q]

                if compute_descriptor:
                    for j in range(128):
                        flat_descriptors[total_keypoints * 128 + j] = \
                            min(512.0 * single_descriptor_arr_t[j], 255.0)

                total_keypoints += 1
            i += 1

    if verbose:
        printf("vl_sift: found %d keypoints\n", total_keypoints)

    # cleanup
    vl_sift_delete(filt)

    # If we have dynamically allocated memory for the frames, make sure that
    # we resize the array back to the correct size (since we optimistically
    # allocated previously to reduce the number of total resizes)
    if not user_specified_frames:
        out_frames = np.resize(out_frames, (total_keypoints, 4))
        out_descriptors = np.resize(out_descriptors, (total_keypoints, 128))

    if compute_descriptor:
        if float_descriptors:
            return out_frames, out_descriptors
        else:
            return out_frames, out_descriptors.astype(np.uint8)
    else:
        return out_frames



import numpy as np
from .cysift import cy_sift


def sift(image, n_octaves=None, n_levels=3,  first_octave=0,  peak_thresh=0,
         edge_thresh=10, norm_thresh=None,  magnification=3, window_size=2,
         frames=None, force_orientations=False, float_descriptors=False,
         compute_descriptor=False, verbose=False):
    r"""
    Extracts a set of SIFT features from ``image``. ``image`` must be
    ``float32`` and greyscale (either a single channel as the last axis, or no
    channel). Each column of ``frames`` is a feature frame and has the format
    ``[Y, X, S, TH]``, where ``(Y, X)`` is the floating point center of the
    keypoint, ``S`` is the scale and ``TH`` is the orientation (in radians).

    If ``compute_descriptors=True``, computes the SIFT descriptors as well. Each
    column of ``descriptors`` is the descriptor of the corresponding frame in
    ``frames``. A descriptor is a 128-dimensional vector of type ``uint8``.

    Parameters
    ----------
    image : [H, W] or [H, W, 1] `float32` `ndarray`
        A single channel, greyscale, `float32` numpy array (ndarray)
        representing the image to calculate descriptors for.
    n_octaves : `int`, optional
        The number of octaves of the DoG scale space. If ``None``, the maximum
        number of octaves will be used.
    n_levels : `int`, optional
        The number of levels per octave of the DoG scale space.
    first_octave : `int`, optional
        The index of the first octave of the DoG scale space.
    peak_thresh : `int`, optional
        The peak selection threshold. The peak threshold filters peaks of the
        DoG scale space that are too small (in absolute value).
    edge_thresh : `int`, optional
        The edge selection threshold. The edge threshold eliminates peaks of the
        DoG scale space whose curvature is too small (such peaks yield badly
        localized frames).
    norm_thresh : `float`, optional
        Set the minimum l2-norm of the descriptors before normalization.
        Descriptors below the threshold are set to zero. If ``None``,
        norm_thresh is ``-inf``.
    magnification : `int`, optional
        Set the descriptor magnification factor. The scale of the keypoint is
        multiplied by this factor to obtain the width (in pixels) of the spatial
        bins. For instance, if there are there are 4 spatial bins along each
        spatial direction, the ``side`` of the descriptor is approximately ``4 *
        magnification``.
    window_size : `int`, optional
        Set the variance of the Gaussian window that determines the
        descriptor support. It is expressed in units of spatial bins.
    frames : `[F, 4]` `float32` `ndarray`, optional
        If specified, set the frames to use (bypass the detector). If frames are
        not passed in order of increasing scale, they are re-orderded. A frame
        is a vector of length 4 ``[Y, X, S, TH]``, representing a disk of center
        f[:2], scale f[2] and orientation f[3].
    force_orientations : `bool`, optional
        If ``True``, compute the orientations of the frames, overriding the
        orientation specified by the ``frames`` argument.
    float_descriptors : `bool`, optional
        If ``True``, the descriptor are returned in floating point rather than
        integer format.
    compute_descriptor : `bool`, optional
        If ``True``, the descriptors are also returned, as well as the keypoints
        (frames). This means that the output of calling this function changes
        from a single value ``frames``, to a tuple of output values ``(frames,
        descriptors)``.
    verbose : `bool`, optional
        If ``True``, be verbose.

    Returns
    -------
    frames : `(F, 4)` `float32` `ndarray`
        ``F`` is the number of keypoints (frames) used. This is the center
        of every dense SIFT descriptor that is extracted.
    descriptors : `(F, 128)` `uint8` or `float32` `ndarray`, optional
        ``F`` is the number of keypoints (frames) used. The 128 length vectors
        per keypoint extracted. ``uint8`` by default. Only returned if
        ``compute_descriptors=True``.
    """
    # Remove last channel
    if image.ndim == 3 and image.shape[-1] == 1:
        image = image[..., 0]

    # Validate image size
    if image.ndim != 2:
        raise ValueError('Only 2D arrays are supported')

    if frames is not None:
        if frames.ndim != 2 or frames.shape[-1] != 4:
            raise ValueError('Frames should be a 2D array of size '
                             '(n_keypoints, 4)')
        frames = np.require(frames, dtype=np.float32, requirements='C')

    # Validate all the parameters
    if n_octaves is not None and n_octaves < 0:
        raise ValueError('n_octaves must be >= 0')
    if n_octaves is None:
        n_octaves = -1
    if n_levels < 1:
        raise ValueError('n_levels must be > 0')
    if first_octave < 0:
        raise ValueError('first_octave must be >= 0')
    if edge_thresh < 1:
        raise ValueError('edge_thresh must be > 0')
    if peak_thresh < 0:
        raise ValueError('peak_thresh must be >= 0')
    if norm_thresh is not None and norm_thresh < 0:
        raise ValueError('norm_thresh must be >= 0')
    if norm_thresh is None:
        norm_thresh = -1
    if window_size < 0:
        raise ValueError('window_size must be >= 0')

    # Ensure types are correct before passing to Cython
    image = np.require(image, dtype=np.float32, requirements='C')

    result = cy_sift(image, n_octaves, n_levels,
                     first_octave,  peak_thresh,
                     edge_thresh, norm_thresh,  magnification,
                     window_size, frames, force_orientations,
                     float_descriptors, compute_descriptor,
                     verbose)
    # May be a tuple or a single return of only the calculated frames
    return result



import numpy as np
from .cysift import cy_dsift


def dsift(image, step=1, size=3, bounds=None, window_size=-1, norm=False,
          fast=False, float_descriptors=False, geometry=(4, 4, 8),
          verbose=False):
    r"""
    Extracts a dense set of SIFT features from ``image``. ``image`` must be
    ``float32`` and greyscale (either a single channel as the last axis, or no
    channel).

    **Important!:** Unlike the Matlab wrapper of vlfeat, this method always
    takes the input arguments as Y-axis, then X-axis. For example, the step
    [2, 3] denotes 2 pixels in the Y-axis, then two pixels in the X-axis. This
    is to create a 1-1 mapping from the input parameters to the shape of the
    image, which is always denoted [H, W] (size of Y-domain, size of X-domain).
    Also, note that the frames are not 1-based as in Matlab, and therefore
    the top left corner is ``(0, 0)`` not ``(1, 1)``.

    Does NOT compute a Gaussian scale space of the image. Instead, the image
    should be pre-smoothed at the desired scale level.

    The scale of the extracted descriptors is controlled by the option
    ``size``, i.e. the width in pixels of a spatial bin (recall that a
    SIFT descriptor is a spatial histogram with 4 x 4 bins).

    The sampling density is controlled by the option ``step``, which is
    the horizontal and vertical displacement of each feature center to
    the next.

    The sampled image area is controlled by the option ``bounds``,
    defining a rectangle in which features are computed. A descriptor
    is included in the rectangle if all the centers of the spatial
    bins are included. The upper-left descriptor is placed so that the
    upper-left spatial bin center is aligned with the upper-left
    corner of the rectangle.

    By default, ``dsift`` computes features equivalent to
    ``sift``. However, the ``fast`` option can be used to turn on an
    variant of the descriptor which, while not strictly equivalent, is much
    faster.

    **Relation to SIFT**

    In the standard SIFT detector/descriptor, implemented by
    ``sift``, the size of a spatial bin is related to the keypoint
    scale by a multiplier, called magnification factor, and denoted
    `magnification`. Therefore, the keypoint scale corresponding to the
    descriptors extracted by ``dsift`` is equal to ``size`` /
    ``magnification``. ``dsift`` does not use ``magnification`` because, by
    using dense sampling, it avoids detecting keypoints in the first place.

    ``dsift`` does not smooth the image as SIFT does. Therefore, in
    order to obtain equivalent results, the image should be
    pre-smoothed appropriately. Recall that in SIFT, for a keypoint of
    scale ``S``, the image is pre-smoothed by a Gaussian of variance
    ``S.^2 - 1/4``.

    **Further details on the geometry**

    As mentioned, the ``dsift`` descriptors cover the bounding box
    specified by ``bounds = [YMIN, XMIN, YMAX, XMAX]``. Thus the top-left bin
    of the top-left descriptor is placed at ``(YMIN, XMIN)``. The next
    three bins to the right are at ``YMIN + size``, ``YMIN + 2*size``,
    ``YMIN + 3*size``. The Y coordinate of the center of the first descriptor is
    therefore at ``(YMIN + YMIN + 3*size) / 2 = YMIN + 3/2 * size``. For
    instance, if ``YMIN=1`` and ``size=3`` (default values), the Y
    coordinate of the center of the first descriptor is at
    ``1 + 3/2 * 3 = 5.5``. For the second descriptor immediately to its right
    this is ``5.5 + step``, and so on.

    Parameters
    ----------
    image : [H, W] or [H, W, 1] `float32` `ndarray`
        A single channel, greyscale, `float32` numpy array (ndarray)
        representing the image to calculate descriptors for.
    step : `int`, optional
        A SIFT descriptor is extracted every ``step`` pixels. This allows for
        sub-sampling of the image.
    size : `int`, optional
        The size of the spatial bin of the SIFT descriptor in pixels.
    bounds : [`int`, `int`, `int`, `int`], optional
        Specifies a rectangular area where descriptors should be
        extracted. The format is ``[YMIN, XMIN, YMAX, XMAX]``. If this
        option is not specified, the entire image is used. The
        bounding box is clipped to the image boundaries.
    norm : `bool`, optional
        If ``True``, adds to the ``frames`` output argument a third
        row containing the descriptor norm, or energy, before
        contrast normalization. This information can be used to
        suppress low contrast descriptors.
    fast : `bool`, optional
        If ``True``, use a piecewise-flat, rather than Gaussian,
        windowing function. While this breaks exact SIFT equivalence,
        in practice is much faster to compute.
    float_descriptors : `bool`, optional
        If ``True``, the descriptor are returned in floating point
        rather than integer format.
    geometry : [`int`, `int`, `int`], optional
        Specify the geometry of the descriptor as ``[NY, NX, NO]``, where ``NY``
        is the number of bins in the Y direction, NX in the X direction,
        and NO the number of orientation bins.
    verbose : `bool`, optional
        If ``True``, be verbose.

    Returns
    -------
    frames : `(F, 2)` or `(F, 3)` `float32` `ndarray`
        ``F`` is the number of keypoints (frames) used. This is the center
        of every dense SIFT descriptor that is extracted.
    descriptors : `(F, 128)` `uint8` or `float32` `ndarray`
        ``F`` is the number of keypoints (frames) used. The 128 length vectors
        per keypoint extracted. ``uint8`` by default.
    """
    # Remove last channel
    if image.ndim == 3 and image.shape[-1] == 1:
        image = image[..., 0]

    # Validate image size
    if image.ndim != 2:
        raise ValueError('Only 2D arrays are supported')

    # Validate bounds
    if bounds is None:
        bounds = np.array([0, 0, image.shape[0] - 1, image.shape[1] - 1])
    else:
        bounds = np.asarray(bounds)
    if bounds.shape[0] != 4:
        raise ValueError('Bounds must be contain 4 elements.')
    for b in bounds:
        if b < 0:
            raise ValueError('Bounds must only contain integers greater than '
                             'or equal to 0.')
    if bounds[2] > image.shape[0] - 1 or bounds[3] > image.shape[1] - 1:
        raise ValueError('Bounds must be the size of the image or less.')

    # Validate size
    if isinstance(size, int):
        size = np.array([size, size])
    else:
        size = np.asarray(size)
    if size.shape[0] != 2:
        raise ValueError('Size vector must contain exactly 2 elements.')
    for s in size:
        if s < 1:
            raise ValueError('Size must only contain positive integers.')

    # Validate step
    if isinstance(step, int):
        step = np.array([step, step])
    else:
        step = np.asarray(step)
    if step.shape[0] != 2:
        raise ValueError('Step vector must contain exactly 2 elements.')
    for s in step:
        if s < 1:
            raise ValueError('Step must only contain positive integers.')

    # Validate window_size
    if not isinstance(window_size, int):
        raise ValueError('Window size must be an integer.')

    # Validate geometry
    geometry = np.asarray(geometry)
    if geometry.shape[0] != 3:
        raise ValueError('Geometry must contain exactly 3 integer elements.')
    if np.min(geometry) < 1:
        raise ValueError('Geometry must only contain positive integers.')

    geometry = geometry.astype(np.int32)
    step = step.astype(np.int32)
    size = size.astype(np.int32)
    bounds = bounds.astype(np.int32)
    image = np.require(image, dtype=np.float32, requirements='C')
    frames, descriptors = cy_dsift(image, step, size, bounds, window_size,
                                   norm, fast, float_descriptors, geometry,
                                   verbose)
    return frames, descriptors



import numpy as np
from .cyfisher import cy_fisher


def fisher(x, means, covariances, priors, normalized=False, square_root=False,
           improved=False, fast=False, verbose=False):
    r"""
    Computes the Fisher vector encoding of the vectors ``x`` relative to
    the diagonal covariance Gaussian mixture model with ``means``,
    ``covariances``, and prior mode probabilities ``priors``.

    By default, the standard Fisher vector is computed.

    Parameters
    ----------
    x : [D, N]  `float32` `ndarray`
        One column per data vector (e.g. a SIFT descriptor)
    means :  [F, N]  `float32` `ndarray`
        One column per GMM component.
    covariances :  [F, N]  `float32` `ndarray`
        One column per GMM component (covariance matrices are assumed diagonal,
        hence these are simply the variance of each data dimension).
    priors :  [F, N]  `float32` `ndarray`
        Equal to the number of GMM components.
    normalized : `bool`, optional
        If ``True``, L2 normalize the Fisher vector.
    square_root : `bool`, optional
        If ``True``, the signed square root function is applied to the return
        vector before normalization.
    improved : `bool`, optional
        If ``True``, compute the improved variant of the Fisher Vector. This is
        equivalent to specifying the ``normalized`` and ``square_root` options.
    fast : `bool`, optional
        If ``True``, uses slightly less accurate computations but significantly
        increase the speed in some cases (particularly with a large number of
        Gaussian modes).
    verbose: `bool`, optional
        If ``True``, print information.

    Returns
    -------
    enc : [k, 1] `float32` `ndarray`
        A vector of size equal to the product of
        ``k = 2 * the n_data_dimensions * n_components``.
    """
    # check for None
    if x is None or means is None or covariances is None or priors is None:
        raise ValueError('A required input is None')

    # validate the gmm parameters
    D = means.shape[0]  # the feature dimensionality
    K = means.shape[1]  # the number of GMM modes
    # N = x.shape[1] is the number of samples
    if covariances.shape[0] != D:
        raise ValueError('covariances and means do not have the same '
                         'dimensionality')
    
    if priors.ndim != 1:
        raise ValueError('priors has an unexpected shape')
    
    if covariances.shape[1] != K or priors.shape[0] != K:
        raise ValueError('covariances or priors does not have the correct '
                         'number of modes')

    if x.shape[0] != D:
        raise ValueError('x and means do not have the same dimensionality')

    return cy_fisher(x, means, covariances, priors,
                     np.int32(normalized), np.int32(square_root),
                     np.int32(improved), np.int32(fast), np.int32(verbose))



# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# Copyright (C) 2013 Andrea Vedaldi and David Novotny.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
from .host cimport vl_size, vl_index, vl_bool, vl_type, vl_uint32
from .mathop cimport VlFloatVectorComparisonFunction, VlDoubleVectorComparisonFunction, VlVectorComparisonType


cdef extern from "vl/kmeans.h":
    
    cdef enum _VlKMeansAlgorithm:
        VlKMeansLloyd = 0,       # Lloyd algorithm
        VlKMeansElkan = 1,       # Elkan algorithm
        VlKMeansANN = 2          # Approximate nearest neighbors
    ctypedef _VlKMeansAlgorithm VlKMeansAlgorithm

    cdef enum _VlKMeansInitialization:
        VlKMeansRandomSelection = 0,  # Randomized selection
        VlKMeansPlusPlus = 1          # Plus plus raondomized selection
    ctypedef _VlKMeansInitialization VlKMeansInitialization

    cdef struct _VlKMeans:
        vl_type dataType                      # Data type.
        vl_size dimension                     # Data dimensionality.
        vl_size numCenters                    # Number of centers.
        vl_size numTrees                      # Number of trees in forest when using ANN-kmeans.
        vl_size maxNumComparisons             # Maximum number of comparisons when using ANN-kmeans.

        VlKMeansInitialization initialization # Initalization algorithm.
        VlKMeansAlgorithm algorithm           # Clustring algorithm.
        VlVectorComparisonType distance       # Distance.
        vl_size maxNumIterations              # Maximum number of refinement iterations.
        double minEnergyVariation             # Minimum energy variation.
        vl_size numRepetitions                # Number of clustering repetitions.
        int verbosity                         # Verbosity level.

        void * centers                        # Centers
        void * centerDistances                # Centers inter-distances.

        double energy                         # Current solution energy.
        VlFloatVectorComparisonFunction floatVectorComparisonFn
        VlDoubleVectorComparisonFunction doubleVectorComparisonFn
    ctypedef _VlKMeans VlKMeans

    VlKMeans * vl_kmeans_new(vl_type dataType, VlVectorComparisonType distance)
    VlKMeans * vl_kmeans_new_copy(VlKMeans* kmeans)
    void vl_kmeans_delete(VlKMeans * self)


    void vl_kmeans_reset(VlKMeans * self)

    double vl_kmeans_cluster(VlKMeans * self,
                            void* data,
                            vl_size dimension,
                            vl_size numData,
                            vl_size numCenters)

    void vl_kmeans_quantize(VlKMeans * self,
                           vl_uint32 * assignments,
                           void * distances,
                           void* data,
                           vl_size numData)

    void vl_kmeans_quantize_ANN(VlKMeans * self,
                               vl_uint32 * assignments,
                               void * distances,
                               void* data,
                               vl_size numData,
                               vl_size iteration )

    void vl_kmeans_set_centers(VlKMeans * self,
                              void* centers,
                              vl_size dimension,
                              vl_size numCenters)

    void vl_kmeans_init_centers_with_rand_data(VlKMeans * self,
                                               void* data,
                                               vl_size dimensions,
                                               vl_size numData,
                                               vl_size numCenters)

    void vl_kmeans_init_centers_plus_plus(VlKMeans * self,
                                           void* data,
                                           vl_size dimensions,
                                           vl_size numData,
                                           vl_size numCenters)

    double vl_kmeans_refine_centers(VlKMeans * self,
                                   void* data,
                                   vl_size numData)

    inline vl_type vl_kmeans_get_data_type (VlKMeans* self)
    inline VlVectorComparisonType vl_kmeans_get_distance (VlKMeans* self)

    inline VlKMeansAlgorithm vl_kmeans_get_algorithm (VlKMeans* self)
    inline VlKMeansInitialization vl_kmeans_get_initialization (VlKMeans* self)
    inline vl_size vl_kmeans_get_num_repetitions (VlKMeans* self)

    inline vl_size vl_kmeans_get_dimension (VlKMeans* self)
    inline vl_size vl_kmeans_get_num_centers (VlKMeans* self)

    inline int vl_kmeans_get_verbosity (VlKMeans* self)
    inline vl_size vl_kmeans_get_max_num_iterations (VlKMeans* self)
    inline double vl_kmeans_get_min_energy_variation (VlKMeans* self)
    inline vl_size vl_kmeans_get_max_num_comparisons (VlKMeans* self)
    inline vl_size vl_kmeans_get_num_trees (VlKMeans* self)
    inline double vl_kmeans_get_energy (VlKMeans* self)
    inline void* vl_kmeans_get_centers (VlKMeans* self)
    
    inline void vl_kmeans_set_algorithm (VlKMeans * self, VlKMeansAlgorithm algorithm)
    inline void vl_kmeans_set_initialization (VlKMeans * self, VlKMeansInitialization initialization)
    inline void vl_kmeans_set_num_repetitions (VlKMeans * self, vl_size numRepetitions)
    inline void vl_kmeans_set_max_num_iterations (VlKMeans * self, vl_size maxNumIterations)
    inline void vl_kmeans_set_min_energy_variation (VlKMeans * self, double minEnergyVariation)
    inline void vl_kmeans_set_verbosity (VlKMeans * self, int verbosity)
    inline void vl_kmeans_set_max_num_comparisons (VlKMeans * self, vl_size maxNumComparisons)
    inline void vl_kmeans_set_num_trees (VlKMeans * self, vl_size numTrees)



from __future__ import division
from cyvlfeat.kmeans import kmeans
import numpy as np


def same_set(values1, values2, precision):
    """
    Compare two sets of values and check that they are roughly the same
    Parameters
    ----------
    values1 : NxD set of values
    values2 : MxD set of values
    precision : precision allowed for the values to mismatch

    Returns
    -------
    Whether or not each set contains a neighbouring pair in the other set
    """
    dimension = values1.shape[1]
    assert values2.shape[1] == dimension
    dist = np.sqrt(np.sum(
        np.power(values1.reshape((-1, 1, dimension)) - values2.reshape((1, -1, dimension)), 2),
        axis=2))*1/dimension
    assert isinstance(dist, np.ndarray)
    assert dist.shape == (values1.shape[0], values2.shape[1])
    return np.all(np.min(dist, axis=0) < precision) and np.all(np.min(dist, axis=1) < precision)


def test_kmeans_float():
    num_data = 50
    num_centers = 4
    dimension = 4
    noise_level = 0.1

    centers = np.random.random_integers(-20, 20, (num_centers, dimension)).astype(np.float32)
    data = np.empty((num_data, dimension), dtype=np.float32)
    for i in range(num_data):
        data[i] = centers[i % num_centers] + np.random.random_sample(dimension)*noise_level

    found_centers, found_assignments = kmeans(data, num_centers, initialization="PLUSPLUS")

    assert found_centers.dtype == np.float32
    assert found_centers.shape == (num_centers, dimension)

    assert same_set(centers, found_centers, 0.1)

    for i in range(num_centers):
        for j in range(num_centers):
            if i != j:
                assert found_assignments[i] != found_assignments[j]
    for i in range(num_data):
        assert found_assignments[i] == found_assignments[i % num_centers]


def test_kmeans_double():
    num_data = 50
    num_centers = 4
    dimension = 4
    noise_level = 0.1

    centers = np.random.random_integers(-20, 20, (num_centers, dimension)).astype(np.float64)
    data = np.empty((num_data, dimension), dtype=np.float64)
    for i in range(num_data):
        data[i] = centers[i % num_centers] + np.random.random_sample(dimension)*noise_level

    found_centers, found_assignments = kmeans(data, num_centers, initialization="PLUSPLUS")

    assert found_centers.dtype == np.float64
    assert found_centers.shape == (num_centers, dimension)

    assert same_set(centers, found_centers, 0.1)

    for i in range(num_centers):
        for j in range(num_centers):
            if i != j:
                assert found_assignments[i] != found_assignments[j]
    for i in range(num_data):
        assert found_assignments[i] == found_assignments[i % num_centers]


def test_kmeans_ANN():
    num_data = 5000
    num_centers = 4
    dimension = 4
    noise_level = 0.1

    centers = np.random.random_integers(-20, 20, (num_centers, dimension)).astype(np.float32)
    data = np.empty((num_data, dimension), dtype=np.float32)
    for i in range(num_data):
        data[i] = centers[i % num_centers] + np.random.random_sample(dimension)*noise_level

    found_centers, found_assignments = kmeans(data, num_centers, initialization="PLUSPLUS", algorithm="ANN")

    assert found_centers.dtype == np.float32
    assert found_centers.shape == (num_centers, dimension)

    assert same_set(centers, found_centers, 0.1)

    for i in range(num_centers):
        for j in range(num_centers):
            if i != j:
                assert found_assignments[i] != found_assignments[j]
    for i in range(num_data):
        assert found_assignments[i] == found_assignments[i % num_centers]



# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
from .host cimport vl_size

cdef extern from "vl/mathop.h":
    double VL_PI
    ctypedef float (*VlFloatVectorComparisonFunction)(vl_size dimension, const float* X, const float* Y)
    ctypedef double (*VlDoubleVectorComparisonFunction)(vl_size dimension, const double* X, const double* Y)
    
    cdef enum _VlVectorComparisonType:
        VlDistanceL1 = 0,        # l1 distance (squared intersection metric)
        VlDistanceL2 = 1,        # squared l2 distance
        VlDistanceChi2 = 2,      # squared Chi2 distance
        VlDistanceHellinger = 3, # squared Hellinger's distance
        VlDistanceJS = 4,        # squared Jensen-Shannon distance
        VlDistanceMahalanobis = 5,     # squared mahalanobis distance
        VlKernelL1 = 6,          # intersection kernel
        VlKernelL2 = 7,          # l2 kernel
        VlKernelChi2 = 8,        # Chi2 kernel
        VlKernelHellinger = 9,   # Hellinger's kernel
        VlKernelJS = 10          # Jensen-Shannon kernel
    ctypedef _VlVectorComparisonType VlVectorComparisonType

    inline char* vl_get_vector_comparison_type_name(int type)


import numpy as np
from .cykmeans import cy_kmeans, algorithm_type, initialization_type, distance_type


def kmeans(data, num_centers, distance="l2", initialization="RANDSEL",
           algorithm="LLOYD", num_repetitions=1, num_trees=3, max_num_comparisons=100,
           max_num_iterations=100, min_energy_variation=0, verbose=False):
    """
    Clusters ``data`` using the K-means algorithm. ``data`` can be either np.float32 or np.foat64. Returns the centers
    of the clusters as an np.ndarray of the same dtype as ``data``. Also returns the cluster assignments
    Parameters
    ----------
    data : [N,D] np.ndarray of dtype in [np.float32, np.float64]
        Input data to be clustered.
    num_centers : `int`
        Number of clusters to compute
    distance : `str`, optional
        Distance to be used ["l1", "l2"]. Default : "l2"
    initialization : `str`, optional
        Use either random data points "RANDSEL" or k-means++ "PLUSPLUS"
        to initialize the centers. Default : "RANDSEL"
    algorithm : `str`, optional
        One of "LLOYD", "ELKAN", or "ANN". LLOYD is the standard Lloyd
        algorithm (similar to expectation maximisation). ELKAN is a
        faster version of LLOYD using triangular inequalities to cut
        down significantly the number of sample-to-center
        comparisons. ANN is the same as Lloyd, but uses an approximated
        nearest neighbours (ANN) algorithm to accelerate the
        sample-to-center comparisons. The latter is particularly
        suitable for very large problems. Default : "LLOYD"
    num_repetitions : `int`, optional
        Number of time to restart k-means. The solution with minimal
        energy is returned. Default : 1
    num_trees : `int`, optional
        The number of trees int the randomized KD-Tree forest (for "ANN"). Default : 3
    max_num_comparisons : `int`, optional
        Maximum number of sample-to-center comparisons when searching
        for the closest center (for "ANN"). Default : 100
    max_num_iterations : `int`, optional
        Maximum number of iterations allowed for the kmeans algorithm
        to converge (for "ANN"). Default : 100
    min_energy_variation : `float`, optional
        Default : 0 (no effect)
    verbose : `bool`, optional
        Outputs information of the converging process

    Returns
    -------
    (centers, assignments) : ([num_centers,D] np.ndarray of same dtype as ``data``, [N] `np.uint32` np.ndarray)
        Computed clusters and the respective assignments of the data points.
    """

    assert isinstance(data, np.ndarray)
    assert isinstance(num_centers, int)
    assert isinstance(verbose, bool)
    if data.ndim != 2:
        raise ValueError('Data should be a 2-D matrix')

    if data.dtype not in [np.float32, np.float64]:
        raise ValueError('Data should be float32 or float64')

    if num_centers > data.shape[0]:
        raise ValueError('num_centers should be a positive integer smaller than the number of data points')

    distance_b = distance.encode()
    if distance_b not in distance_type.keys():
        raise ValueError('distance field invalid')

    initialization_b = initialization.encode()
    if initialization_b not in initialization_type.keys():
        raise ValueError('initialization field invalid')

    algorithm_b = algorithm.encode()
    if algorithm_b not in algorithm_type.keys():
        raise ValueError('algorithm field invalid')

    if (not isinstance(num_repetitions, int)) or num_repetitions <= 0:
        raise ValueError('num_repetitions should be a positive integer')
    if (not isinstance(num_trees, int)) or num_trees <= 0:
        raise ValueError('num_trees should be a positive integer')
    if (not isinstance(max_num_comparisons, int)) or max_num_comparisons <= 0:
        raise ValueError('max_num_comparisons should be a positive integer')
    if (not isinstance(max_num_iterations, int)) or max_num_iterations <= 0:
        raise ValueError('max_num_iterations should be a positive integer')

    return cy_kmeans(data, num_centers, distance_b, initialization_b,
                     algorithm_b, num_repetitions, num_trees, max_num_comparisons,
                     max_num_iterations, min_energy_variation, verbose)



# distutils: language = c
# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
import numpy as np
cimport numpy as np
cimport cython
from cyvlfeat._vl.host cimport *
from cyvlfeat._vl.mathop cimport *
from cyvlfeat._vl.kmeans cimport *
from libc.stdio cimport printf
from libc.string cimport memcpy

algorithm_type = {b'LLOYD': VlKMeansLloyd, b'ELKAN': VlKMeansElkan, b'ANN': VlKMeansANN}

initialization_type = {b'RANDSEL': VlKMeansRandomSelection, b'PLUSPLUS': VlKMeansPlusPlus}

distance_type = {b'l1': VlDistanceL1, b'l2': VlDistanceL2}

cpdef cy_kmeans(np.ndarray data, int num_centers, bytes distance, bytes initialization,
           bytes algorithm, int num_repetitions, int num_trees, int max_num_comparisons,
           int max_num_iterations, int min_energy_variation, bint verbose):

    cdef:
        VlKMeans* kmeans
        double energy
        int dimension
        int num_data
        vl_type data_type
        np.ndarray centers
        np.ndarray[unsigned int, ndim=1, mode='c'] assignments
        np.uint8_t[:,:] data_view = data.view(np.uint8)
        np.uint8_t[:,:] centers_view

    dimension = data.shape[1]
    num_data = data.shape[0]

    if data.dtype == np.float32:
        data_type = VL_TYPE_FLOAT
    else:
        data_type = VL_TYPE_DOUBLE
    kmeans = vl_kmeans_new(data_type, distance_type[distance])

    vl_kmeans_set_verbosity(kmeans, verbose)
    vl_kmeans_set_num_repetitions(kmeans, num_repetitions)
    vl_kmeans_set_algorithm(kmeans, algorithm_type[algorithm])
    vl_kmeans_set_initialization(kmeans, initialization_type[initialization])
    vl_kmeans_set_max_num_iterations(kmeans, max_num_iterations)
    vl_kmeans_set_max_num_comparisons(kmeans, max_num_comparisons)
    vl_kmeans_set_num_trees(kmeans, num_trees)

    if min_energy_variation>0:
        vl_kmeans_set_min_energy_variation(kmeans, min_energy_variation)
        
    if verbose:
        printf("kmeans: Initialization = %s\n", initialization)
        printf("kmeans: Algorithm = %s\n", algorithm)
        printf("kmeans: MaxNumIterations = %d\n", vl_kmeans_get_max_num_iterations(kmeans))
        printf("kmeans: MinEnergyVariation = %f\n", vl_kmeans_get_min_energy_variation(kmeans))
        printf("kmeans: NumRepetitions = %d\n", vl_kmeans_get_num_repetitions(kmeans))
        printf("kmeans: data type = %s\n", vl_get_type_name(vl_kmeans_get_data_type(kmeans)))
        printf("kmeans: distance = %s\n", vl_get_vector_comparison_type_name(vl_kmeans_get_distance(kmeans)))
        printf("kmeans: data dimension = %d\n", dimension)
        printf("kmeans: num. data points = %d\n", num_data)
        printf("kmeans: num. centers = %d\n", num_centers)
        printf("kmeans: max num. comparisons = %d\n", max_num_comparisons)
        printf("kmeans: num. trees = %d\n", num_trees)
        printf("\n")

    energy = vl_kmeans_cluster(kmeans, &data_view[0, 0], dimension, num_data, num_centers)

    if data_type == VL_TYPE_FLOAT:
        centers = np.empty((num_centers, dimension), dtype=np.float32, order='C')
    else:
        centers = np.empty((num_centers, dimension), dtype=np.float64, order='C')
    centers_view = centers.view(np.uint8)

    assignments = np.empty((num_data,), dtype=np.uint32, order='C')

    memcpy(&centers_view[0,0], vl_kmeans_get_centers (kmeans),
          vl_get_type_size(data_type) * dimension * vl_kmeans_get_num_centers(kmeans))

    vl_kmeans_quantize(kmeans, &assignments[0], NULL, &data_view[0, 0], num_data)

    vl_kmeans_delete(kmeans)

    return centers, assignments
